import colorsys
import re
from collections import OrderedDict, defaultdict
from functools import partial

import pandas as pd
import numpy as np

from fireant.dataset.fields import (
    DataType,
    Field,
)
from fireant.dataset.filters import ComparisonOperator
from fireant.dataset.totals import (
    DATE_TOTALS,
    NUMBER_TOTALS,
    TEXT_TOTALS,
    TOTALS_MARKERS,
)
from fireant.formats import (
    RAW_VALUE,
    TOTALS_VALUE,
    display_value,
    json_value,
    raw_value,
    return_none,
    safe_value,
)
from fireant.reference_helpers import reference_alias
from fireant.utils import (
    alias_for_alias_selector,
    alias_selector,
    getdeepattr,
    setdeepattr,
    wrap_list,
)
from .base import ReferenceItem
from .pandas import Pandas

TOTALS_LABEL = "Totals"
METRICS_DIMENSION_ALIAS = "metrics"
F_METRICS_DIMENSION_ALIAS = alias_selector(METRICS_DIMENSION_ALIAS)
_display_value = partial(display_value, nan_value="", null_value="")


def hex_to_rgb(hex_val):
    """
    https://stackoverflow.com/questions/29643352/converting-hex-to-rgb-value-in-python
    """
    return tuple(int(hex_val[i:i + 2], 16) for i in (0, 2, 4))


def rgb_to_hex(rgb_val):
    """
    https://stackoverflow.com/questions/3380726/converting-a-rgb-color-tuple-to-a-six-digit-code-in-python
    """
    return "{0:02x}{1:02x}{2:02x}".format(*rgb_val)


def hex_to_hsv(hex_color):
    rgb_color = hex_to_rgb(hex_color)
    return colorsys.rgb_to_hsv(*(val / 255 for val in rgb_color))


class FormattingField:
    def __init__(self, metric=None, reference=None, operation=None):
        self.metric = metric
        self.reference = reference
        self.operation = operation

    def get_alias(self):
        if self.reference:
            if self.operation:
                return reference_alias(self.operation, self.reference)
            else:
                return reference_alias(self.metric, self.reference)

        if self.operation:
            return self.operation.alias

        return self.metric.alias


class FormattingRule:
    def __init__(self, field, color, covers_row):
        self.field = field
        self.color = color
        self.covers_row = covers_row

    def applies(self, value):
        return True

    def determine_colors(self, value):
        background_color = self._determine_background_color(value)
        return background_color, self._determine_text_color(background_color)

    def _determine_background_color(self, value):
        return self.color

    def _determine_text_color(self, background_color):
        """
        Determines, using the luminance of the background color, whether the text color should be light or dark.
        """
        rgb_bg_color = hex_to_rgb(background_color)
        luminance = (0.299 * rgb_bg_color[0] + 0.587 * rgb_bg_color[1] + 0.114 * rgb_bg_color[2]) / 255;
        return 'FDFDFD' if luminance < 0.5 else '212121';

    def get_field_selector(self):
        return alias_selector(self.field.get_alias())


class FormattingConditionRule(FormattingRule):
    def __init__(self, field, operator, value, color, covers_row=False):
        super().__init__(field, color, covers_row)
        self.operator = operator
        self.value = value

    def applies(self, value):
        return ComparisonOperator.eval(value, self.operator, self.value)


class FormattingHeatMapRule(FormattingRule):
    """
    The colors of the heatmap are calculated using the HSV color format.
    The Hue (position in the color spectrum) and Value (brightness) of the final color are the same as the rule color.
    The Saturation (color purity) of the final color depends on both the submitted value and the saturation of the rule
    color. We calculate a saturation_spread which is either the saturation of the rule minus 0.05 or 0.50, whatever is
    highest. This we will multiply with the ratio of the value compared to the min and max values which should be set
    with set_min_max. The final value we increase with 0.05 to not have any white values.
    This will result in for instance a final saturation range of [0.05; 1.00] if the rule color is very saturated
    and in [0.05; 0.55] for a low saturated rule color.

    If a start_color is supplied the same calculations are done for both the color and the start_color, but the
    start_color will cover the first 50% of values and it will result in a color transition over white.
    """

    WHITE = 'FFFFFF'

    def __init__(self, field, color, start_color=None, covers_row=False):
        super().__init__(field, color, covers_row)
        self.hsv_color = hex_to_hsv(self.color)
        if start_color:
            self.hsv_start_color = hex_to_hsv(start_color)
            self.start_saturation_spread = max(0.50, self.hsv_start_color[1])
        else:
            self.hsv_start_color = None
        self.saturation_spread = max(0.50, self.hsv_color[1] - 0.05)
        self.min_val, self.max_val = None, None

    def set_min_max(self, min_val, max_val):
        self.min_val, self.max_val = min_val, max_val

    @staticmethod
    def _is_invalid(val):
        return pd.isna(val) or np.isinf(val)

    @staticmethod
    def get_hex_color_with_new_saturation(base_hsv_color, saturation):
        calculated_hsv_color = colorsys.hsv_to_rgb(base_hsv_color[0], saturation, base_hsv_color[2])
        return rgb_to_hex((round(val * 255) for val in calculated_hsv_color))

    def _determine_background_color(self, value):
        if self._is_invalid(value) or self._is_invalid(self.min_val) or self._is_invalid(self.max_val):
            return self.WHITE

        val_ratio = (value - self.min_val) / (self.max_val - self.min_val)

        if self.hsv_start_color is not None:
            if val_ratio <= 0.5:
                # The supplied value should use the start color.
                # Move the val_ratio back to a [0, 1.0] range with the values flipped.
                flipped_val_ratio = abs(val_ratio - 0.5) * 2.0
                saturation = flipped_val_ratio * self.start_saturation_spread
                return self.get_hex_color_with_new_saturation(self.hsv_start_color, saturation)
            else:
                # The supplied value should use the default color.
                # Adjust it from the (0.5, 1.0] range to (0.0, 1.0]
                val_ratio = (val_ratio - 0.5) * 2.0

        saturation = val_ratio * self.saturation_spread + 0.05
        return self.get_hex_color_with_new_saturation(self.hsv_color, saturation)


def find_rule_to_apply(rules, value):
    for rule in rules:
        if rule.applies(value):
            return rule

    return None


def map_index_level(index, level, func):
    # If the index is empty, do not do anything
    if 0 == index.size:
        return index

    if isinstance(index, pd.MultiIndex):
        values = index.levels[level]
        return index.set_levels(values.map(func), level)

    assert level == 0

    return index.map(func)


class TotalsItem:
    alias = TOTALS_VALUE
    label = TOTALS_LABEL
    prefix = suffix = precision = None


class ReactTable(Pandas):
    """
    This component does not work with react-table out of the box, some customization is needed in order to work with
    the transformed data.

    .. code-block:: jsx

        // A Custom TdComponent implementation is required by Fireant in order to render display values
        const TdComponent = ({
                               toggleSort,
                               className,
                               children,
                               ...rest
                             }) =>
            <div className={classNames('rt-td', className)} role="gridcell" {...rest}>
                {_.get(children, 'display', children.raw) || <span>&nbsp;</span>}
            </div>;

        const FireantReactTable = ({
                                config, // The payload from fireant
                              }) =>
            <ReactTable columns={config.columns}
                        data={config.data}
                        minRows={0}

                        TdComponent={ DashmoreTdComponent}
                        defaultSortMethod={(a, b, desc) => ReactTableDefaults.defaultSortMethod(a.raw, b.raw, desc)}>
            </ReactTable>;
    """
    def __init__(
        self,
        metric,
        *metrics: Field,
        pivot=(),
        hide=(),
        transpose=False,
        sort=None,
        ascending=None,
        max_columns=None,
        formatting_rules=(),
    ):
        super(ReactTable, self).__init__(
            metric,
            *metrics,
            pivot=pivot,
            hide=hide,
            transpose=transpose,
            sort=sort,
            ascending=ascending,
            max_columns=max_columns
        )
        self.formatting_rules_map = defaultdict(list)
        self.min_max_map = {}
        for formatting_rule in formatting_rules:
            field_selector = formatting_rule.get_field_selector()
            self.formatting_rules_map[field_selector].append(formatting_rule)
            if isinstance(formatting_rule, FormattingHeatMapRule):
                self.min_max_map[field_selector] = [np.inf, -np.inf]

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__, ",".join(str(m) for m in self.items)
        )

    @staticmethod
    def map_hyperlink_templates(df, dimensions):
        """
        Creates a mapping for each dimension to it's hyperlink template if it is possible to create the hyperlink
        template for it.

        The hyperlink template is a URL-like string containing curley braces enclosing dimension keys: `{dimension}`.
        While rendering this widget, the dimension key placeholders need to be replaced with the dimension values for
        that row.

        :param df:
            The result data set that is being transformed. The data frame SHOULD be pivoted/transposed if that step is
            required, before calling this function, in order to prevent the template from being included for the
            dimension if one of the required dimensions is pivoted.
        :param dimensions:
            The list of dimensions included in the query that created the result data set df.
        :return:
            A dict with the dimension key as the key and the hyperlink template as the value. Templates will only be
            included if it will be possible to fill in the required parameters.
        """
        hyperlink_templates = {}
        pattern = re.compile(r"{[^{}]+}")

        for dimension in dimensions:
            hyperlink_template = dimension.hyperlink_template
            if hyperlink_template is None:
                continue

            required_hyperlink_parameters = [
                alias_selector(argument[1:-1])
                for argument in pattern.findall(hyperlink_template)
            ]

            # Check that all of the required dimensions are in the result data set. Only include the hyperlink template
            # in the return value of this function if all are present.
            unavailable_hyperlink_parameters = set(required_hyperlink_parameters) & set(
                df.index.names
            )
            if not unavailable_hyperlink_parameters:
                continue

            # replace the dimension keys with the formatted values. This will come in handy later when replacing the
            # actual values
            hyperlink_template = hyperlink_template.format(
                **{
                    alias_for_alias_selector(argument): "{" + argument + "}"
                    for argument in required_hyperlink_parameters
                }
            )

            f_dimension_alias = alias_selector(dimension.alias)
            hyperlink_templates[f_dimension_alias] = hyperlink_template

        return hyperlink_templates

    @staticmethod
    def format_data_frame(data_frame):
        """
        This function prepares the raw data frame for transformation by formatting dates in the index and removing any
        remaining NaN/NaT values. It also names the column as metrics so that it can be treated like a dimension level.

        :param data_frame:
            The result set data frame
        :param dimensions:
        :return:
        """
        data_frame = data_frame.copy()
        data_frame.columns.name = F_METRICS_DIMENSION_ALIAS
        return data_frame

    @staticmethod
    def transform_index_column_headers(data_frame, field_map, hide_dimension_aliases):
        """
        Convert the un-pivoted dimensions into ReactTable column header definitions.

        :param data_frame:
            The result set data frame.
        :param field_map:
            A map to find metrics/operations based on their keys found in the data frame.
        :param hide_dimension_aliases:
            A set with hide dimension aliases.
        :return:
            A list of column header definitions with the following structure.

        .. code-block:: jsx

            columns = [{
              Header: 'Column A',
              accessor: 'a',
            }, {
              Header: 'Column B',
              accessor: 'b',
            }]
        """
        columns = []
        if (
            not isinstance(data_frame.index, pd.MultiIndex)
            and data_frame.index.name is None
        ):
            return columns

        for f_dimension_alias in data_frame.index.names:
            if f_dimension_alias not in field_map or f_dimension_alias in hide_dimension_aliases:
                continue

            dimension = field_map[f_dimension_alias]
            header = getattr(dimension, "label", dimension.alias)

            columns.append(
                {
                    "Header": json_value(header),
                    "accessor": safe_value(f_dimension_alias),
                }
            )

        return columns

    @staticmethod
    def transform_data_column_headers(data_frame, field_map):
        """
        Convert the metrics into ReactTable column header definitions. This includes any pivoted dimensions, which will
        result in multiple rows of headers.

        :return:
        :param data_frame:
            The result set data frame.
        :param field_map:
            A map to find metrics/operations based on their keys found in the data frame.
        :return:
            A list of column header definitions with the following structure.

        .. code-block:: jsx

            columns = [{
              Header: 'Column A',
              columns: [{
                Header: 'SubColumn A.0',
                accessor: 'a.0',
              }, {
                Header: 'SubColumn A.1',
                accessor: 'a.1',
              }]
            }, {
              Header: 'Column B',
              columns: [
                ...
              ]
            }]
        """

        def get_header(column_value, f_dimension_alias, is_totals):
            if f_dimension_alias == F_METRICS_DIMENSION_ALIAS or is_totals:
                item = field_map[column_value]
                return getattr(item, "label", item.alias)

            if f_dimension_alias in field_map:
                field = field_map[f_dimension_alias]
                return _display_value(column_value, field) or safe_value(column_value)

            if f_dimension_alias is None:
                return ""

            return safe_value(column_value)

        def _make_columns(columns_frame, previous_level_values=()):
            """
            This function recursively creates the individual column definitions for React Table with the above tree
            structure depending on how many index levels there are in the columns.

            :param columns_frame:
                A data frame representing the columns of the result set data frame.
            :param previous_level_values:
                A tuple containing the higher level index level values used for building the data accessor path
            """
            f_dimension_alias = columns_frame.index.names[0]

            # Group the columns if they are multi-index so we can get the proper sub-column values. This will yield
            # one group per dimension value with the group data frame containing only the relevant sub-columns
            groups = (
                columns_frame.groupby(level=0)
                if isinstance(columns_frame.index, pd.MultiIndex)
                else [(level, None) for level in columns_frame.index]
            )

            columns = []
            for column_value, group in groups:
                is_totals = column_value in TOTALS_MARKERS | {TOTALS_LABEL}

                # All column definitions have a header
                column = {
                    "Header": get_header(column_value, f_dimension_alias, is_totals)
                }

                level_values = previous_level_values + (column_value,)
                if group is not None:
                    # If there is a group, then drop this index level from the group data frame and recurse to build
                    # sub column definitions
                    next_level_df = group.reset_index(level=0, drop=True)
                    column["columns"] = _make_columns(next_level_df, level_values)

                else:
                    column["accessor"] = ".".join(
                        safe_value(value) for value in level_values
                    )

                if is_totals:
                    column["className"] = "fireant-totals"

                columns.append(column)

            return columns

        # If the query only has a single metric, that level will be dropped, and set as data_frame.name
        dropped_metric_level_name = (
            (data_frame.name,) if hasattr(data_frame, "name") else ()
        )

        return _make_columns(data_frame.columns.to_frame(), dropped_metric_level_name)

    @staticmethod
    def transform_row_index(index_values, field_map, dimension_hyperlink_templates, hide_dimension_aliases, row_colors):
        # Add the index to the row
        row = {}
        for key, value in index_values.items():
            if key is None or key not in field_map:
                continue

            field_alias = key
            field = field_map[field_alias]

            data = {RAW_VALUE: raw_value(value, field)}
            display = _display_value(value, field)
            if display is not None:
                data["display"] = display
            if row_colors is not None:
                data["color"], data["text_color"] = row_colors

            is_totals = display == TOTALS_LABEL

            # If the dimension has a hyperlink template, then apply the template by formatting it with the dimension
            # values for this row. The values contained in `index_values` will always contain all of the required values
            # at this point, otherwise the hyperlink template will not be included.
            if not is_totals and key in dimension_hyperlink_templates:
                try:
                    data["hyperlink"] = dimension_hyperlink_templates[key].format(
                        **index_values
                    )
                except KeyError:
                    pass

            safe_key = safe_value(key)
            row[safe_key] = data

        for dimension_alias in hide_dimension_aliases:
            del row[dimension_alias]

        return row

    @staticmethod
    def _get_row_value_accessor(series, fields, key):
        index_names = series.index.names or []

        accessor_fields = [
            fields[field_alias]
            for field_alias in index_names
            if field_alias is not None
        ]
        accessor = [
            safe_value(value) for value, field in zip(key, accessor_fields)
        ] or key

        return accessor

    def transform_row_values(self, series, fields, is_transposed, is_pivoted):
        row = {}
        row_colors = None

        for key, value in series.items():
            key = wrap_list(key)

            # Get the field for the metric
            metric_alias = wrap_list(series.name)[0] if is_transposed else key[0]
            field = fields[metric_alias]
            data = {
                RAW_VALUE: raw_value(value, field),
            }
            if not row_colors:
                # No color for this field yet
                rule = find_rule_to_apply(self.formatting_rules_map[metric_alias], value)
                if rule is not None:
                    colors = rule.determine_colors(value)
                    data["color"], data["text_color"] = colors
                    if not is_transposed and not is_pivoted and rule.covers_row:
                        # No transposing or pivoting going on so set as row color if it's specified for the rule
                        row_colors = colors

            display = _display_value(value, field, date_as=return_none)
            if display is not None:
                data["display"] = display

            accessor = self._get_row_value_accessor(series, fields, key)
            setdeepattr(row, accessor, data)

        # Assign the row colors to fields that aren't colored yet
        if row_colors:
            for key in series.keys():
                accessor = self._get_row_value_accessor(series, fields, wrap_list(key))
                data = getdeepattr(row, accessor)
                if "color" not in data:
                    data["color"], data["text_color"] = row_colors

        return row, row_colors

    def calculate_min_max(self, df, is_transposed):
        if not self.min_max_map:
            return

        for index, series in df.iterrows():
            for key, value in series.items():
                metric_alias = wrap_list(series.name)[0] if is_transposed else wrap_list(key)[0]
                min_max = self.min_max_map.get(metric_alias)
                if min_max is not None:  # we need to calculate the min and max values for this metric.
                    if value < min_max[0]:
                        min_max[0] = value  # value is smaller than stored smallest value, replace it.
                    if value > min_max[1]:
                        min_max[1] = value  # value is bigger than stored biggest value, replace it.

        for rule_list in self.formatting_rules_map.values():
            for rule in rule_list:
                if isinstance(rule, FormattingHeatMapRule):
                    rule.set_min_max(*self.min_max_map[rule.get_field_selector()])

    def transform_data(
        self, data_frame, field_map, hide_aliases, dimension_hyperlink_templates, is_transposed, is_pivoted,
    ):
        """
        Builds a list of dicts containing the data for ReactTable. This aligns with the accessors set by
        #transform_dimension_column_headers and #transform_metric_column_headers

        :param data_frame:
            The result set data frame.
        :param field_map:
            A mapping to all the fields in the dataset used for this query.
        :param hide_aliases:
            A set with hide dimension aliases.
        :param dimension_hyperlink_templates:
            A mapping to fields and its hyperlink dimension, if any.
        :param is_transposed:
            Whether the table is transposed or not.
        :param is_pivoted:
            Whether the table is pivoted or not.
        """
        index_names = data_frame.index.names

        def _get_field_label(alias):
            if alias not in field_map:
                return alias

            field = field_map[alias]
            return getattr(field, "label", field.alias)

        # If the metric column was dropped due to only having a single metric, add it back here so the
        # formatting can be applied.
        if hasattr(data_frame, "name"):
            metric_alias = data_frame.name
            data_frame = pd.concat(
                [data_frame],
                keys=[metric_alias],
                names=[F_METRICS_DIMENSION_ALIAS],
                axis=1,
            )

        self.calculate_min_max(data_frame, is_transposed)

        rows = []
        for index, series in data_frame.iterrows():
            row_values, row_colors = self.transform_row_values(series, field_map, is_transposed, is_pivoted)

            index = wrap_list(index)
            # Get a list of values from the index. These can be metrics or dimensions so it checks in the item map if
            # there is a display value for the value
            index_values = (
                [_get_field_label(value) for value in index] if is_transposed else index
            )
            index_display_values = OrderedDict(zip(index_names, index_values))
            row_index = self.transform_row_index(
                index_display_values, field_map, dimension_hyperlink_templates, hide_aliases, row_colors
            )
            rows.append(
                {
                    **row_index,
                    **row_values,
                }
            )

        return rows

    def transform(
        self,
        data_frame,
        dimensions,
        references,
        annotation_frame=None,
        use_raw_values=False,
    ):
        """
        Transforms a data frame into a format for ReactTable. This is an object containing attributes `columns` and
        `data` which align with the props in ReactTable with the same name.

        :param data_frame:
            The result set data frame.
        :param dimensions:
            A list of dimensions that were selected in the data query.
        :param references:
            A list of references that were selected in the data query.
        :param annotation_frame:
            A data frame containing the annotation data.
        :param use_raw_values:
            Don't add prefix or postfix to values.
        :return:
            An dict containing attributes `columns` and `data` which align with the props in ReactTable with the same
            names.
        """
        result_df = data_frame.copy()

        dimension_map = {
            alias_selector(dimension.alias): dimension for dimension in dimensions
        }

        metric_map = OrderedDict(
            [
                (
                    alias_selector(reference_alias(item, ref)),
                    ReferenceItem(item, ref) if ref is not None else item,
                )
                for item in self.items
                for ref in [None] + references
            ]
        )

        field_map = {
            **metric_map,
            **dimension_map,
            # Add an extra item to map the totals markers to it's label
            NUMBER_TOTALS: TotalsItem,
            TEXT_TOTALS: TotalsItem,
            DATE_TOTALS: TotalsItem,
            TOTALS_LABEL: TotalsItem,
            alias_selector(METRICS_DIMENSION_ALIAS): Field(
                METRICS_DIMENSION_ALIAS, None, data_type=DataType.text, label=""
            ),
        }
        metric_aliases = list(metric_map.keys())

        hide_aliases = {
            alias_selector(dimension.alias) for dimension in self.hide
        }

        for dimension in dimensions:
            if dimension.fetch_only:
                hide_aliases.add(alias_selector(dimension.alias))

        pivot_dimensions = [
            alias_selector(dimension.alias)
            for dimension in self.pivot
            if alias_selector(dimension.alias) not in hide_aliases
        ]

        result_df = self.format_data_frame(result_df[metric_aliases])
        result_df, is_pivoted, is_transposed = self.pivot_data_frame(result_df, pivot_dimensions, self.transpose)
        dimension_columns = self.transform_index_column_headers(result_df, field_map, hide_aliases)
        metric_columns = self.transform_data_column_headers(result_df, field_map)

        data = self.transform_data(
            result_df,
            field_map,
            hide_aliases=hide_aliases,
            dimension_hyperlink_templates=self.map_hyperlink_templates(result_df, dimensions),
            is_transposed=is_transposed,
            is_pivoted=is_pivoted,
        )

        return {"columns": dimension_columns + metric_columns, "data": data}



import codecs
import os
import re

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


with open("requirements.txt") as file:
    install_requires = file.read().splitlines()

extras_requires = {}
extras_requires_patterns = re.compile(r"requirements-extras-(\w+)\.txt")
for file_name in os.listdir("."):
    match = extras_requires_patterns.search(file_name)
    if not match:
        continue

    with open(file_name) as file:
        extras_name = match.group(1)
        extras_requires[extras_name] = file.read().splitlines()

setup(
    name="fireant",
    version=find_version("fireant", "__init__.py"),
    author="KAYAK, GmbH",
    author_email="bandit@kayak.com",
    # License
    license="Apache License Version 2.0",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    # Include additional files into the package
    include_package_data=True,
    # Details
    url="https://github.com/kayak/fireant",
    description="Data analysis tool for Python and Jupyter Notebooks",
    long_description="Fireant is a tool which aims to simplify the process of accessing analytical data and building "
    "dashboards and reports.  It eliminates the need for handwritten queries or code to process "
    "data. "
    "The main feature, the dataset, is a schema that is used to configure metrics, the data being "
    "visualized, and dimensions, the spatiality over which metrics can be extended and grouped. "
    "Included in the Slicer is a Slicer Manager API which exposes methods to render tables and "
    "charts "
    "in a variety of different formats.\n\n"
    "Fireant is a great compliment to Jupter notebooks using matplotlib and pandas. It can be used "
    "in "
    "web projects with Highcharts and React-Table.  It can even just provide a layer of "
    "abstraction "
    "for accessing analytical data akin to ORM libraries in software projects.",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: PL/SQL",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    keywords=(
        "fireant python query builder querybuilder sql mysql postgres mssql psql oracle vertica aggregated "
        "relational database rdbms business analytics bi data science analysis pandas"
    ),
    install_requires=install_requires,
    extras_require=extras_requires,
    tests_require=["mock"],
    test_suite="fireant.tests",
)



from datetime import datetime

from pypika import CustomFunction, MSSQLQuery, Parameter, Table
from pypika.functions import Cast, DateDiff
from pypika.terms import Function, PseudoColumn

from .base import Database

_MSSQLDateAdd = CustomFunction('DATEADD', ['date_part', 'number', 'term'])
_MSSQLConvert = CustomFunction('CONVERT', ['term', 'expression', 'style'])


class MSSQLDatabase(Database):
    """
    Microsoft SQL Server client that uses the pymssql module under the hood.
    """

    # The pypika query class to use for constructing queries
    query_cls = MSSQLQuery
    can_group_static_value = False

    def __init__(
        self,
        host='localhost',
        port=1433,
        database=None,
        user=None,
        password=None,
        **kwargs
    ):
        super().__init__(host, port, database, **kwargs)
        self.user = user
        self.password = password

    def connect(self):
        import pymssql
        return pymssql.connect(
            server=self.host,
            port=self.port,
            database=self.database,
            user=self.user,
            password=self.password,
        )

    def trunc_date(self, field, interval):
        # Useful docs on this here: http://www.silota.com/docs/recipes/sql-server-date-parts-truncation.html
        return self.date_add(0, interval, DateDiff(PseudoColumn(interval), 0, field))

    def date_add(self, field, date_part, interval):
        return _MSSQLDateAdd(PseudoColumn(date_part), interval, field)

    def convert_date(self, dt: datetime) -> Function:
        return Cast(dt, 'datetimeoffset')

    def get_column_definitions(self, schema, table, connection=None):
        columns = Table('COLUMNS', schema='INFORMATION_SCHEMA')

        columns_query = (
            MSSQLQuery.from_(columns, immutable=False)
            .select(columns.COLUMN_NAME, columns.DATA_TYPE)
            .where(columns.TABLE_SCHEMA == Parameter('%(schema)s'))
            .where(columns.field('TABLE_NAME') == Parameter('%(table)s'))
            .distinct()
            .orderby(columns.column_name)
        )

        return self.fetch(str(columns_query), connection=connection, parameters=dict(schema=schema, table=table))



import re


class TypeEngine:
    """
    Base class for database type conversions.
    """
    def __init__(self, db_to_ansi_mapper, ansi_to_db_mapper):
        self.db_to_ansi_mapper = db_to_ansi_mapper
        self.ansi_to_db_mapper = ansi_to_db_mapper

    def to_ansi(self, data_type):
        """
        Translates the provided data type string into an ANSI data type instance.

        :param data_type: The data type string to be translated.
        :return: An instance of an ANSI data type.
        """
        raw_data_type, raw_arguments = self.split_data_type(data_type.lower())

        if raw_data_type not in self.db_to_ansi_mapper:
            raise ValueError('Could not find matching ANSI type for {}.'.format(raw_data_type))

        ansi_data_type = self.db_to_ansi_mapper[raw_data_type]

        return ansi_data_type(*raw_arguments)

    def from_ansi(self, data_type):
        """
        Translates an ANSI data type instance to a database specific representation.

        :param data_type: The data type to be translated.
        :return: A string representation of the database specific data type.
        """
        if data_type not in self.ansi_to_db_mapper:
            raise ValueError('Could not find matching database type for {}.'.format(str(data_type)))

        return self.ansi_to_db_mapper[data_type] + data_type.params

    @staticmethod
    def split_data_type(data_type):
        """
        Splits up a data type string into type name and type arguments.

        :param data_type: The data type string to be split up.
        :return: The raw data type and raw arguments.
        """
        split_data_type = re.findall('\w+', data_type)

        raw_data_type = split_data_type[0] if split_data_type else ''
        raw_arguments = split_data_type[1:] if len(split_data_type) > 1 else []

        return raw_data_type, raw_arguments



import pandas as pd

from fireant.dataset.fields import Field
from fireant.utils import (
    alias_for_alias_selector,
    immutable,
)
from .query_builder import QueryBuilder, QueryException, add_hints
from ..execution import fetch_data
from ..sql_transformer import make_latest_query


class DimensionLatestQueryBuilder(QueryBuilder):
    def __init__(self, dataset):
        super().__init__(dataset)

    @immutable
    def __call__(self, dimension: Field, *dimensions: Field):
        self._dimensions += [dimension] + list(dimensions)

    @property
    def sql(self):
        """
        Serializes this query builder as a set of SQL queries.  This method will always return a list of one
        query since only one query is required to retrieve dimension choices.

        This function only handles dimensions (select+group by) and filtering (where/having), which is everything
        needed for the query to fetch choices for dimensions.

        The dataset query extends this with metrics, references, and totals.
        """
        if not self.dimensions:
            raise QueryException(
                "Must select at least one dimension to query latest values"
            )

        query = make_latest_query(
            database=self.dataset.database,
            base_table=self.table,
            joins=self.dataset.joins,
            dimensions=self.dimensions,
        )
        return [query]

    def fetch(self, hint=None):
        queries = add_hints(self.sql, hint)
        max_rows_returned, data = fetch_data(self.dataset.database, queries, self.dimensions)
        data = self._get_latest_data_from_df(data)
        return self._transform_for_return(data, max_rows_returned=max_rows_returned)

    def _get_latest_data_from_df(self, df: pd.DataFrame) -> pd.Series:
        latest = df.reset_index().iloc[0]
        # Remove the row index as the name and trim the special dimension key characters from the dimension key
        latest.name = None
        latest.index = [alias_for_alias_selector(alias) for alias in latest.index]
        return latest



from pypika import (
    EmptyCriterion,
    Not,
)
from pypika.functions import Lower

from .modifiers import FieldModifier


class ComparisonOperator:
    eq = "eq"
    ne = "ne"
    gt = "gt"
    lt = "lt"
    gte = "gte"
    lte = "lte"

    mapping = {
        '=': eq,
        '!=': ne,
        '>': gt,
        '<': lt,
        '>=': gte,
        '<=': lte,
    }

    @staticmethod
    def eval(value_1, operator, value_2):
        if operator == ComparisonOperator.eq:
            return value_1 == value_2

        if operator == ComparisonOperator.ne:
            return value_1 != value_2

        if operator == ComparisonOperator.lt:
            return value_1 < value_2

        if operator == ComparisonOperator.lte:
            return value_1 <= value_2

        if operator == ComparisonOperator.gt:
            return value_1 > value_2

        if operator == ComparisonOperator.gte:
            return value_1 >= value_2


class Filter(FieldModifier):
    @property
    def definition(self):
        raise NotImplementedError()

    @property
    def is_aggregate(self):
        return self.definition.is_aggregate

    @property
    def is_wrapped(self) -> bool:
        """
        We do not wrap field modifiers e.g. filters, just fields themselves during data blending.
        This is for convenience as Filter and Field classes are often passed to util functions as part of the
        same collection.
        """
        return False

    def __eq__(self, other):
        return all(
            [
                isinstance(other, self.__class__),
                str(self.definition) == str(other.definition),
            ]
        )

    def __repr__(self):
        return str(self.definition)


class ComparatorFilter(Filter):

    def __init__(self, field, operator, value):
        self.operator = operator
        self.value = value
        super(ComparatorFilter, self).__init__(field)

    @property
    def definition(self):
        return getattr(self.field.definition, self.operator)(self.value)

    @property
    def is_aggregate(self):
        return self.field.is_aggregate


class BooleanFilter(Filter):
    def __init__(self, field, value):
        self.value = value
        super(BooleanFilter, self).__init__(field)

    @property
    def definition(self):
        if self.value:
            return self.field.definition
        return Not(self.field.definition)


class ContainsFilter(Filter):
    def __init__(self, field, values):
        self.values = values
        super(ContainsFilter, self).__init__(field)

    @property
    def definition(self):
        return self.field.definition.isin(self.values)


class ExcludesFilter(ContainsFilter):
    @property
    def definition(self):
        return self.field.definition.notin(self.values)


class RangeFilter(Filter):
    def __init__(self, field, start, stop):
        self.start = start
        self.stop = stop
        super(RangeFilter, self).__init__(field)

    @property
    def definition(self):
        return self.field.definition.between(self.start, self.stop)


class PatternFilter(Filter):
    def __init__(self, field, pattern, *patterns):
        self.patterns = (pattern, *patterns)
        super(PatternFilter, self).__init__(field)

    @property
    def definition(self):
        first, *rest = self.patterns
        definition = Lower(self.field.definition).like(Lower(first))

        for pattern in rest:
            definition |= Lower(self.field.definition).like(Lower(pattern))

        return definition


class NegatedFilterMixin:
    @property
    def definition(self):
        definition = super().definition
        return definition.negate()


class AntiPatternFilter(NegatedFilterMixin, PatternFilter):
    pass


class VoidFilter(Filter):
    def __init__(self, field):
        super(VoidFilter, self).__init__(field)

    @property
    def definition(self):
        return EmptyCriterion()

    def __repr__(self):
        return "VoidFilter()"



from typing import Iterable, List, Sequence, Type

from pypika import Table, functions as fn
from pypika.queries import QueryBuilder

from fireant.database import Database
from fireant.dataset.fields import Field
from fireant.dataset.filters import Filter
from fireant.dataset.joins import Join
from fireant.dataset.modifiers import Rollup
from fireant.utils import (
    alias_selector,
    flatten,
)
from .field_helper import make_term_for_field
from .finders import (
    find_and_group_references_for_dimensions,
    find_joins_for_tables,
    find_required_tables_to_join,
    find_totals_dimensions,
)
from .references import adapt_for_reference_query
from .special_cases import apply_special_cases
from .totals_helper import adapt_for_totals_query


@apply_special_cases
def make_slicer_query_with_totals_and_references(
    *,
    database,
    table,
    joins,
    dimensions,
    metrics,
    operations,
    filters,
    references,
    orders,
    share_dimensions=(),
) -> List[Type[QueryBuilder]]:
    """
    :param dataset:
    :param database:
    :param table:
    :param joins:
    :param dimensions:
    :param metrics:
    :param operations:
    :param filters:
    :param references:
    :param orders:
    :param share_dimensions:
    :return:
    """

    """
    The following two loops will run over the spread of the two sets including a NULL value in each set:
     - reference group (WoW, MoM, etc.)
     - dimension with roll up/totals enabled (totals dimension)

    This will result in at least one query where the reference group and totals dimension is NULL, which shall be
    called base query. The base query will ALWAYS be present, even if there are zero reference groups or totals
    dimensions.

    For a concrete example, check the test case in :
    ```
    fireant.tests.queries.test_build_dimensions.QueryBuilderDimensionTotalsTests
        #test_build_query_with_totals_cat_dimension_with_references
    ```
    """
    totals_dimensions = find_totals_dimensions(
        dimensions, share_dimensions,
    )
    totals_dimensions_and_none = [None] + totals_dimensions[::-1]

    reference_groups = find_and_group_references_for_dimensions(
        dimensions, references
    )
    reference_groups_and_none = [(None, None)] + list(reference_groups.items())

    queries = []
    for totals_dimension in totals_dimensions_and_none:
        (dimensions_with_totals, filters_with_totals) = adapt_for_totals_query(
            totals_dimension, dimensions, filters,
        )

        for reference_parts, references in reference_groups_and_none:
            (
                dimensions_with_ref,
                metrics_with_ref,
                filters_with_ref,
            ) = adapt_for_reference_query(
                reference_parts,
                database,
                dimensions_with_totals,
                metrics,
                filters_with_totals,
                references,
            )
            query = make_slicer_query(
                database,
                table,
                joins,
                dimensions_with_ref,
                metrics_with_ref,
                filters_with_ref,
                orders,
            )

            # Add these to the query instance so when the data frames are joined together, the correct references and
            # totals can be applied when combining the separate result set from each query.
            query._totals = totals_dimension
            query._references = references
            queries.append(query)

    return queries


def make_slicer_query(
    database: Database,
    base_table: Table,
    joins: Sequence[Join] = (),
    dimensions: Sequence[Field] = (),
    metrics: Sequence[Field] = (),
    filters: Sequence[Filter] = (),
    orders: Sequence = (),
) -> Type[QueryBuilder]:
    """
    Creates a pypika/SQL query from a list of slicer elements.

    This is the base implementation shared by two implementations: the query to fetch data for a slicer request and
    the query to fetch choices for dimensions.

    This function only handles dimensions (select+group by) and filtering (where/having), which is everything needed
    for the query to fetch choices for dimensions.

    The slicer query extends this with metrics, references, and totals.

    :param database:

    :param base_table:
        pypika.Table - The base table of the query, the one in the FROM clause
    :param joins:
        A collection of joins available in the slicer. This should include all slicer joins. Only joins required for
        the query will be used.
    :param dimensions:
        A collection of dimensions to use in the query.
    :param metrics:
        A collection of metrics to use in the query.
    :param filters:
        A collection of filters to apply to the query.
    :param orders:
        A collection of orders as tuples of the metric/dimension to order by and the direction to order in.

    :return:
    """
    query = database.query_cls.from_(base_table, immutable=False)
    elements = flatten([metrics, dimensions, filters])

    # Add joins
    join_tables_needed_for_query = find_required_tables_to_join(elements, base_table)

    for join in find_joins_for_tables(joins, base_table, join_tables_needed_for_query):
        query = query.join(join.table, how=join.join_type).on(join.criterion)

    # Add dimensions
    for dimension in dimensions:
        dimension_term = make_term_for_field(dimension, database.trunc_date)
        query = query.select(dimension_term)

        # Some database platforms like MSSQL do not support grouping by static value columns.
        # Fireant uses static value columns for totals placeholders.
        # TODO this can be reverted once an issue with data blending attaching unnecessary subqueries
        # is removed as in some cases the left join can cause duplicate rows to appear when not grouping a rollup column
        ungroupable_rollup = isinstance(dimension, Rollup) and not database.can_group_static_value

        if not dimension.is_aggregate and not ungroupable_rollup:
            query = query.groupby(dimension_term)

    # Add filters
    for fltr in filters:
        query = (
            query.having(fltr.definition)
            if fltr.is_aggregate
            else query.where(fltr.definition)
        )

    # Add metrics
    metric_terms = [make_term_for_field(metric) for metric in metrics]
    if metric_terms:
        query = query.select(*metric_terms)

    # In the case that the orders are determined by a field that is not selected as a metric or dimension, then it needs
    # to be added to the query.
    select_aliases = {el.alias for el in query._selects}
    for (orderby_field, orientation) in orders:
        orderby_term = make_term_for_field(orderby_field)
        query = query.orderby(orderby_term, order=orientation)

        if orderby_term.alias not in select_aliases:
            query = query.select(orderby_term)

    return query


def make_latest_query(
    database: Database,
    base_table: Table,
    joins: Iterable[Join] = (),
    dimensions: Iterable[Field] = (),
):
    query = database.query_cls.from_(base_table, immutable=False)

    # Add joins
    join_tables_needed_for_query = find_required_tables_to_join(dimensions, base_table)
    for join in find_joins_for_tables(joins, base_table, join_tables_needed_for_query):
        query = query.join(join.table, how=join.join_type).on(join.criterion)

    for dimension in dimensions:
        f_dimension_key = alias_selector(dimension.alias)
        query = query.select(fn.Max(dimension.definition).as_(f_dimension_key))

    return query



import numpy as np
from fireant import utils
from fireant.utils import immutable
from pypika import functions as fn
from pypika.queries import QueryBuilder

from .modifiers import FieldModifier


class ReferenceFilter:
    def __init__(self, metric, operator, value):
        self.metric = metric
        self.operator = operator
        self.value = value


class Reference(FieldModifier):
    def __init__(self, field, reference_type, delta=False, delta_percent=False, filters=()):
        super().__init__(field)

        self.reference_type = reference_type
        self.alias = (
            reference_type.alias + "_delta_percent"
            if delta_percent
            else reference_type.alias + "_delta"
            if delta
            else reference_type.alias
        )

        self.label = (
            reference_type.label + " Δ%"
            if delta_percent
            else reference_type.label + " Δ"
            if delta
            else reference_type.label
        )

        self.time_unit = reference_type.time_unit
        self.interval = reference_type.interval

        self.delta = delta_percent or delta
        self.delta_percent = delta_percent

        self.filters = filters

    def __eq__(self, other):
        return (
            isinstance(self, Reference)
            and self.field is other.field
            and self.alias == other.alias
        )

    def __hash__(self):
        return hash("reference{}{}".format(self.alias, self.field))

    def __repr__(self):
        return "{}({})".format(self.alias, self.field.alias)

    @immutable
    def for_(self, field):
        self.field = field


class ReferenceType(object):
    def __init__(self, alias, label, time_unit: str, interval: int):
        self.alias = alias
        self.label = label

        self.time_unit = time_unit
        self.interval = interval

    def __call__(self, dimension, delta=False, delta_percent=False, filters=()):
        return Reference(dimension, self, delta=delta, delta_percent=delta_percent, filters=filters)

    def __eq__(self, other):
        return (
            isinstance(self, ReferenceType)
            and self.time_unit == other.time_unit
            and self.interval == other.interval
        )

    def __hash__(self):
        return hash("reference" + self.alias)


DayOverDay = ReferenceType("dod", "DoD", "day", 1)
WeekOverWeek = ReferenceType("wow", "WoW", "week", 1)
MonthOverMonth = ReferenceType("mom", "MoM", "month", 1)
QuarterOverQuarter = ReferenceType("qoq", "QoQ", "quarter", 1)
YearOverYear = ReferenceType("yoy", "YoY", "year", 1)


def reference_term(
    reference: Reference, original_query: QueryBuilder, ref_query: QueryBuilder
):
    """
    Part of query building. Given a reference, the original dataset query, and the ref query, creates the pypika for
    the reference that should be selected in the reference container query.

    :param reference:
    :param original_query:
    :param ref_query:
    :return:
    """

    def original_field(metric):
        return original_query.field(utils.alias_selector(metric.alias))

    def ref_field(metric):
        return ref_query.field(utils.alias_selector(metric.alias))

    if reference.delta:
        if reference.delta_percent:
            return lambda metric: (original_field(metric) - ref_field(metric)) * (
                100 / fn.NullIf(ref_field(metric), 0)
            )

        return lambda metric: original_field(metric) - ref_field(metric)

    return ref_field


def calculate_delta_percent(ref_df, ref_delta_df):
    # pandas raises an exception when dividing by zero
    return 100.0 * ref_delta_df.divide(ref_df.replace(0, np.nan))



import math
from datetime import (
    date,
    datetime,
)

import numpy as np
import pandas as pd

from fireant.dataset.fields import DataType
from fireant.dataset.totals import TOTALS_MARKERS
from fireant.utils import filter_kwargs

RAW_VALUE = "raw"
INF_VALUE = "Inf"
NAN_VALUE = "NaN"
NULL_VALUE = "null"
BLANK_VALUE = ""
TOTALS_VALUE = "$totals"
TOTALS_LABEL = "Totals"

epoch = np.datetime64(datetime.utcfromtimestamp(0))
milliseconds = np.timedelta64(1, "ms")

DATE_FORMATS = {
    "iso": "%Y-%m-%dT%H:%M:%S",
    "hour": "%Y-%m-%d %H:00",
    "day": "%Y-%m-%d",
    "week": "W%W %Y-%m-%d",
    "month": "%b %Y",
    "year": "%Y",
}


def quarter_from_month(month):
    return math.ceil(month / 3)


@filter_kwargs
def date_as_string(value, interval_key=None):
    if interval_key == 'quarter':
        # strftime does not support a quarter format placeholder
        return 'Q{quarter} {year}'.format(year=value.year, quarter=quarter_from_month(value.month))

    f = DATE_FORMATS.get(interval_key, "%Y-%m-%d")
    return value.strftime(f)


@filter_kwargs
def date_as_millis(value):
    if isinstance(value, date) and not isinstance(value, datetime):
        value = datetime.combine(value, datetime.min.time())
    return int(1000 * value.timestamp())


@filter_kwargs
def return_none(value):
    return None


@filter_kwargs
def _format_decimal(value, thousands="", precision=None, suffix=None, use_raw_value=False):
    if not isinstance(value, (int, float)):
        return value

    if use_raw_value and suffix == '%':
        # When raw values are required, we divide percentage values by 100 to ensure they
        # work well with Spreadsheet applications like Excel.
        value /= 100

        # Add extra precision to offset the division
        if precision is not None:
            precision += 2

    if use_raw_value:
        precision_pattern = f'{{:.{precision if precision is not None else 16}f}}'
    elif precision is not None:
        precision_pattern = f'{{:{thousands}.{precision}f}}'
    else:
        precision_pattern = f'{{:{thousands}f}}'

    value = precision_pattern.format(value)
    if precision is None:
        value = value.rstrip('0').rstrip('.')

    return value


@filter_kwargs
def wrap_styling(value, use_raw_value=False, prefix=None, suffix=None):
    if value is None or use_raw_value:
        return value

    return "{prefix}{value}{suffix}".format(
        prefix=prefix or "", suffix=suffix or "", value=value
    )


@filter_kwargs
def _identity(value):
    return value


def json_value(value):
    """
    This function will return only values safe for JSON
    """
    if pd.isnull(value):
        return None
    if isinstance(value, float) and np.isinf(value):
        return "inf"
    if value in TOTALS_MARKERS:
        return TOTALS_VALUE
    return value


def safe_value(value):
    if value is None or value is "" or pd.isnull(value):
        return NULL_VALUE
    if isinstance(value, float) and np.isinf(value):
        return "inf"
    if value in TOTALS_MARKERS:
        return TOTALS_VALUE

    return (
        date_as_string(value, interval_key="iso")
        if isinstance(value, date)
        else str(value)
    )


@filter_kwargs
def _format_date_field(value, date_as=date_as_string, **kwargs):
    return date_as(value, **kwargs)


RAW_FIELD_FORMATTER = {
    DataType.date: _format_date_field,
}


def raw_value(value, field, date_as=date_as_string):
    """
    Converts a raw metric value into a safe type. This will change dates into strings, NaNs into Nones, and np types
    into their corresponding python types.

    :param value:
        The raw metric value.
    :param field:
        The dataset field that the value represents.
    :param date_as:
    """
    if value is None or pd.isnull(value):
        return None
    if isinstance(value, float) and np.isinf(value):
        return None
    if value in TOTALS_MARKERS:
        return TOTALS_VALUE

    formatter = RAW_FIELD_FORMATTER.get(field.data_type, _identity)
    return formatter(value, date_as=date_as, interval_key="iso")


def _format_number_field_value(value, **kwargs):
    number = _format_decimal(value, **kwargs)
    return wrap_styling(number, **kwargs)


@filter_kwargs
def _format_boolean_field(value):
    return str(value).lower()


FIELD_DISPLAY_FORMATTER = {
    DataType.date: _format_date_field,
    DataType.number: _format_number_field_value,
    DataType.boolean: _format_boolean_field,
    DataType.text: return_none,
}


def display_value(
    value, field, date_as=date_as_string, nan_value=NAN_VALUE, null_value=NULL_VALUE, use_raw_value=False,
):
    """
    Converts a metric value into the display value by applying formatting.

    :param value:
        The raw metric value.
    :param field:
        The dataset field that the value represents.
    :param date_as:
        A format function for datetimes.
    :param nan_value:
        The value to return if the value is a Pandas null (np.nan) value.
    :param null_value:
        The value to return if the value is None
    :param use_raw_value:
        Do not output a value with prefix/suffixes. If a suffix is a percentage sign and this value is true, the raw
        value will be value/100 to make it easier for users when exporting to tools like Excel.

    :return:
        A formatted string containing the display value for the metric.
    """
    if value is None:
        return null_value
    if pd.isnull(value):
        return nan_value
    if isinstance(value, float) and np.isinf(value):
        return INF_VALUE
    if value in TOTALS_MARKERS:
        return TOTALS_LABEL

    format_kwargs = {
        key: getattr(field, key, None)
        for key in ("prefix", "suffix", "thousands", "precision", "interval_key")
    }
    format_kwargs = {
        key: value for key, value in format_kwargs.items() if value is not None
    }

    formatter = FIELD_DISPLAY_FORMATTER.get(field.data_type, _identity)
    return formatter(value, date_as=date_as, use_raw_value=use_raw_value, **format_kwargs)



import copy
from functools import partial

from fireant.dataset.fields import Field
from .field_helper import make_term_for_field
from .finders import find_field_in_modified_field


def adapt_for_reference_query(
    reference_parts, database, dimensions, metrics, filters, references
):
    if reference_parts is None:
        return dimensions, metrics, filters

    ref_dim, unit, interval = reference_parts

    trunc = database.trunc_date
    offset_func = partial(database.date_add, date_part=unit, interval=interval)
    offset_func_inv = partial(database.date_add, date_part=unit, interval=-interval)

    ref_dimensions = _make_reference_dimensions(dimensions, ref_dim, offset_func, trunc)
    ref_metrics = _make_reference_metrics(metrics, references[0].reference_type.alias)
    ref_filters = _make_reference_filters(filters, ref_dim, offset_func_inv)

    return ref_dimensions, ref_metrics, ref_filters


def _replace_reference_dimension(dimension, offset_func, trunc_date=None):
    ref_definition = offset_func(make_term_for_field(dimension, trunc_date))
    field = Field(
        alias=dimension.alias,
        definition=ref_definition,
        data_type=dimension.data_type,
        label=dimension.label,
    )

    if hasattr(dimension, "for_"):
        return dimension.for_(field)

    return field


def _make_reference_dimensions(dimensions, ref_dimension, offset_func, trunc_date):
    return [
        _replace_reference_dimension(dimension, offset_func, trunc_date)
        if ref_dimension is find_field_in_modified_field(dimension)
        else dimension
        for dimension in dimensions
    ]


def _make_reference_metrics(metrics, ref_key):
    return [
        Field(
            "{}_{}".format(metric.alias, ref_key),
            metric.definition,
            data_type=metric.data_type,
            label=metric.label,
            prefix=metric.prefix,
            suffix=metric.suffix,
            precision=metric.precision,
        )
        for metric in metrics
    ]


def _make_reference_filters(filters, ref_dimension, offset_func):
    """
    Copies and replaces the reference dimension's definition in all of the filters applied to a dataset query.

    This is used to shift the dimension filters to fit the reference window.

    :param filters:
    :param ref_dimension:
    :param offset_func:
    :return:
    """
    reference_filters = []
    for ref_filter in filters:
        if ref_filter.field is ref_dimension:
            # NOTE: Important to apply the offset function to the start and stop properties because the date math can
            # become expensive over many rows
            ref_filter = copy.copy(ref_filter)
            ref_filter.start = offset_func(ref_filter.start)
            ref_filter.stop = offset_func(ref_filter.stop)

        reference_filters.append(ref_filter)

    return reference_filters



import itertools
from datetime import timedelta

import pandas as pd

from fireant import (
    formats,
    utils,
)
from fireant.dataset.fields import DataType
from fireant.dataset.totals import TOTALS_MARKERS
from fireant.reference_helpers import (
    reference_alias,
    reference_label,
    reference_prefix,
    reference_suffix,
)
from fireant.utils import alias_selector
from .base import TransformableWidget
from .chart_base import (
    ChartWidget,
    ContinuousAxisSeries,
)

DEFAULT_COLORS = (
    "#DDDF0D",
    "#55BF3B",
    "#DF5353",
    "#7798BF",
    "#AAEEEE",
    "#FF0066",
    "#EEAAEE",
    "#DF5353",
    "#7798BF",
    "#AAEEEE",
)

DASH_STYLES = (
    "Solid",
    "Dash",
    "Dot",
    "DashDot",
    "LongDash",
    "LongDashDot",
    "ShortDash",
    "ShortDashDot",
    "LongDashDotDot",
    "ShortDashDotDot",
    "ShortDot",
)

MARKER_SYMBOLS = ("circle", "square", "diamond", "triangle", "triangle-down")

SERIES_NEEDING_MARKER = (ChartWidget.LineSeries, ChartWidget.AreaSeries)
TS_UPPER_BOUND = pd.Timestamp.max - timedelta(seconds=1)
ALWAYS_SHARED_TOOLTIP_CHART_TYPES = {ChartWidget.PieSeries.type}


def has_only_line_series(axis):
    return all([isinstance(series_, ChartWidget.LineSeries) for series_ in axis])


class HighCharts(ChartWidget, TransformableWidget):
    # Pagination should be applied to groups of the 0th index level (the x-axis) in order to paginate series
    group_pagination = True

    def __init__(
        self,
        title=None,
        colors=None,
        x_axis_visible=True,
        tooltip_visible=True,
        split_dimension=None,
    ):
        super(HighCharts, self).__init__()
        self.title = title
        self.colors = colors or DEFAULT_COLORS
        self.x_axis_visible = x_axis_visible
        self.tooltip_visible = tooltip_visible
        self.split_dimension = split_dimension or None

    def __repr__(self):
        return ".".join(["HighCharts()"] + [repr(axis) for axis in self.items])

    def transform(self, data_frame, dimensions, references, annotation_frame=None):
        """
        - Main entry point -

        Transforms a data frame into HighCharts JSON format.

        See https://api.highcharts.com/highcharts/

        :param data_frame:
            The data frame containing the data. Index must match the dimensions parameter.
        :param dimensions:
            A list of dimensions that are being rendered.
        :param references:
            A list of references that are being rendered.
        :param annotation_frame:
            A data frame containing annotation data.
        :return:
            A dict or a list of dicts meant to be dumped as JSON.
        """
        result_df = data_frame.copy()

        hide_dimensions = {
            dimension for dimension in dimensions if dimension.fetch_only
        }
        self.hide_data_frame_indexes(result_df, hide_dimensions)

        dimension_map = {
            alias_selector(dimension.alias): dimension for dimension in dimensions
        }

        render_group = []
        split_dimension = self.split_dimension

        if split_dimension:
            split_dimension_alias = alias_selector(split_dimension.alias)

            categories = self._categories(
                result_df, dimension_map, split_dimension_alias,
            )

            for category in categories:
                render_group.append(
                    [
                        result_df.xs(
                            category, level=split_dimension_alias, drop_level=False
                        ),
                        formats.display_value(category, split_dimension) or category,
                    ]
                )

        if not render_group:
            render_group = [(result_df, None)]

        num_charts = len(render_group)

        charts = [
            self._render_individual_chart(
                chart_df,
                dimensions,
                references,
                annotation_frame=annotation_frame,
                titleSuffix=titleSuffix,
                num_charts=num_charts,
            )
            for chart_df, titleSuffix in render_group
        ]

        return charts[0] if num_charts == 1 else charts

    def _render_individual_chart(
        self,
        data_frame,
        dimensions,
        references,
        annotation_frame=None,
        titleSuffix="",
        num_charts=1,
    ):
        result_df = data_frame

        dimension_map = {
            alias_selector(dimension.alias): dimension for dimension in dimensions
        }

        colors = itertools.cycle(self.colors)

        is_timeseries = dimensions and dimensions[0].data_type == DataType.date

        # Timestamp.max is used as a marker for rolled up dimensions (totals). Filter out the totals value for the
        # dimension used for the x-axis
        if is_timeseries and len(data_frame) > 0:
            result_df = self._remove_date_totals(result_df)

        # Group the results by index levels after the 0th, one for each series
        # This will result in a series for every combination of dimension values and each series will contain a data set
        # across the 0th dimension (used for the x-axis)
        series_data_frames = self._group_by_series(result_df)

        total_num_series = sum([len(axis) for axis in self.items])

        y_axes, series = [], []
        for axis_idx, axis in enumerate(self.items):
            # Tee the colors iterator so we can peek at the next color. The next color in the iterator becomes the
            # axis color. The first series on each axis should match the axis color, and will progress the colors
            # iterator. The next axis color should be the next color in the iterator after the last color used by a
            # series on the current axis in order to get a better variety of color.
            colors, tee_colors = itertools.tee(colors)
            axis_color = next(tee_colors)

            # prepend axes, append series, this keeps everything ordered left-to-right
            y_axes[0:0] = self._render_y_axis(
                axis_idx, axis_color if 1 < total_num_series else None, references
            )

            series += self._render_series(
                axis,
                axis_idx,
                axis_color,
                colors,
                result_df,
                series_data_frames,
                dimensions,
                references,
                is_timeseries,
            )

        categories = self._categories(result_df, dimension_map)
        x_axis = self._render_x_axis(dimensions, categories)

        annotations = []
        if has_only_line_series(axis) and annotation_frame is not None:
            annotations = self._render_annotation(annotation_frame, x_axis)

        extra = {}

        if num_charts > 1:
            extra["title"] = {
                "text": f"{self.title} ({titleSuffix})" if self.title else titleSuffix
            }
            extra["chart"] = {
                # Height of 240px is the smallest we can have, while still fully displaying the chart menu.
                "height": 240
            }

        return {
            "title": {"text": self.title},
            "xAxis": x_axis,
            "yAxis": y_axes,
            "colors": self.colors,
            "series": series,
            "tooltip": {
                "useHTML": True,
                "enabled": self.tooltip_visible,
                # When only a single datapoint per series is available, shared tooltips should be avoided.
                # Since it looks clunky and often supersedes most of the chart.
                "shared": all([
                    len(item['data']) > 1 or item['type'] in ALWAYS_SHARED_TOOLTIP_CHART_TYPES
                    for item in series
                ]),
            },
            "legend": {"useHTML": True},
            "annotations": annotations,
            **extra,
        }

    def _categories(self, data_frame, dimension_map, dimension_alias=None):
        is_mi = isinstance(data_frame.index, pd.MultiIndex)
        levels = data_frame.index.levels if is_mi else [data_frame.index]

        first_level = None

        if dimension_alias:
            level_index = 0

            for i, level in enumerate(levels):
                if level.name == dimension_alias:
                    level_index = i
                    break

            first_level = levels[level_index]
        else:
            first_level = levels[0]

        if first_level is not None and first_level.name is not None:
            dimension_alias = first_level.name
            dimension = dimension_map[dimension_alias]
            return [
                formats.display_value(category, dimension) or category
                if not pd.isnull(category)
                else None
                for category in first_level
            ]

        return []

    def _render_x_axis(self, dimensions, categories):
        """
        Renders the xAxis configuration.

        https://api.highcharts.com/highcharts/xAxis

        :param dimensions:
        :param categories:
        :return:
        """
        is_timeseries = dimensions and dimensions[0].data_type == DataType.date

        if is_timeseries:
            return {"type": "datetime", "visible": self.x_axis_visible}

        return {
            "type": "category",
            "categories": categories if categories else ["All"],
            "visible": self.x_axis_visible,
        }

    def _render_y_axis(self, axis_idx, color, references):
        """
        Renders the yAxis configuration.

        https://api.highcharts.com/highcharts/yAxis

        :param axis_idx:
        :param color:
        :param references:
        :return:
        """
        axis = self.items[axis_idx]

        y_axes = [
            {
                "id": str(axis_idx),
                "title": {"text": None},
                "labels": {"style": {"color": color}},
                "visible": axis.y_axis_visible,
            }
        ]

        y_axes += [
            {
                "id": "{}_{}".format(axis_idx, reference.alias),
                "title": {"text": reference.label},
                "opposite": True,
                "labels": {"style": {"color": color}},
                "visible": axis.y_axis_visible,
            }
            for reference in references
            if reference.delta
        ]

        return y_axes

    def _render_series(
        self,
        axis,
        axis_idx,
        axis_color,
        colors,
        data_frame,
        series_data_frames,
        dimensions,
        references,
        is_timeseries=False,
    ):
        """
        Renders the series configuration.

        https://api.highcharts.com/highcharts/series

        :param axis:
        :param axis_idx:
        :param axis_color:
        :param colors:
        :param series_data_frames:
        :param dimensions:
        :param references:
        :param is_timeseries:
        :return:
        """

        hc_series = []
        for series in axis:
            # Pie Charts, do not break data out into groups, render all data in one pie chart series
            if isinstance(series, self.PieSeries):
                # Pie charts do not group the data up by series so render them separately and then go to the next series
                for reference in [None] + references:
                    hc_series.append(
                        self._render_pie_series(
                            series.metric, reference, data_frame, dimensions
                        )
                    )
                continue

            # For other series types, create a highcharts series for each group (combination of dimension values)

            symbols = itertools.cycle(MARKER_SYMBOLS)
            for (dimension_values, group_df), symbol in zip(
                series_data_frames, symbols
            ):
                dimension_values = utils.wrap_list(dimension_values)
                dimension_label = self._format_dimension_values(
                    dimensions[1:], dimension_values
                )

                hc_series += self._render_highcharts_series(
                    series,
                    group_df,
                    references,
                    dimension_label,
                    is_timeseries,
                    symbol,
                    axis_idx,
                    axis_color,
                    next(colors),
                )

        return hc_series

    def _render_highcharts_series(
        self,
        series,
        series_df,
        references,
        dimension_label,
        is_timeseries,
        symbol,
        axis_idx,
        axis_color,
        series_color,
    ):
        """

        Note on colors:
        - With a single axis, use different colors for each series
        - With multiple axes, use the same color for the entire axis and only change the dash style

        :param series:
        :param series_df:
        :param references:
        :param dimension_label:
        :param is_timeseries:
        :param symbol:
        :param axis_idx:
        :param axis_color:
        :param series_color:
        :return:
        """
        if is_timeseries:
            series_df = series_df.sort_index(level=0)

        results = []
        for reference, dash_style in zip(
            [None] + references, itertools.cycle(DASH_STYLES)
        ):
            field_alias = utils.alias_selector(
                reference_alias(series.metric, reference)
            )
            metric_label = reference_label(series.metric, reference)

            hc_series = {
                "type": series.type,
                "name": "{} ({})".format(metric_label, dimension_label)
                if dimension_label
                else metric_label,
                "data": (
                    self._render_timeseries_data(series_df, field_alias, series.metric)
                    if is_timeseries
                    else self._render_category_data(
                        series_df, field_alias, series.metric
                    )
                ),
                "tooltip": self._render_tooltip(series.metric, reference),
                "yAxis": (
                    "{}_{}".format(axis_idx, reference.alias)
                    if reference is not None and reference.delta
                    else str(axis_idx)
                ),
                "marker": (
                    {"symbol": symbol, "fillColor": axis_color or series_color}
                    if isinstance(series, SERIES_NEEDING_MARKER)
                    else {}
                ),
                "stacking": series.stacking,
            }

            if isinstance(series, ContinuousAxisSeries):
                # Set each series in a continuous series to a specific color
                hc_series["color"] = series_color
                hc_series["dashStyle"] = dash_style

            results.append(hc_series)

        return results

    @staticmethod
    def _render_category_data(group_df, field_alias, metric):
        categories = (
            list(group_df.index.levels[0])
            if isinstance(group_df.index, pd.MultiIndex)
            else list(group_df.index)
        )

        series = []
        for labels, y in group_df[field_alias].iteritems():
            label = labels[0] if isinstance(labels, tuple) else labels
            if pd.isnull(label):
                # ignore nans in index
                continue

            series.append(
                {"x": categories.index(label), "y": formats.raw_value(y, metric)}
            )

        return series

    @staticmethod
    def _render_timeseries_data(group_df, metric_alias, metric):
        series = []
        for dimension_values, y in group_df[metric_alias].iteritems():
            first_dimension_value = utils.wrap_list(dimension_values)[0]

            # Ignore empty result sets where the only row is totals
            if first_dimension_value in TOTALS_MARKERS:
                continue

            if pd.isnull(first_dimension_value):
                # Ignore totals on the x-axis.
                continue

            series.append(
                (
                    formats.date_as_millis(first_dimension_value),
                    formats.raw_value(y, metric),
                )
            )
        return series

    def _render_tooltip(self, metric, reference):
        return {
            "valuePrefix": reference_prefix(metric, reference),
            "valueSuffix": reference_suffix(metric, reference),
            "valueDecimals": metric.precision,
        }

    def _render_pie_series(self, metric, reference, data_frame, dimension_fields):
        metric_alias = utils.alias_selector(metric.alias)

        if self.split_dimension:
            dimension_fields = [
                dimension for dimension in dimension_fields if dimension != self.split_dimension
            ]
            data_frame = data_frame.reset_index(alias_selector(self.split_dimension.alias), drop=True)

        data = []
        for dimension_values, y in data_frame[metric_alias].iteritems():
            dimension_values = utils.wrap_list(dimension_values)
            name = self._format_dimension_values(dimension_fields, dimension_values)

            data.append(
                {"name": name or metric.label, "y": formats.raw_value(y, metric)}
            )

        return {
            "name": reference_label(metric, reference),
            "type": "pie",
            "data": data,
            "tooltip": {
                "pointFormat": '<span style="color:{point.color}">\u25CF</span> {series.name}: '
                "<b>{point.y} ({point.percentage:.1f}%)</b><br/>",
                "valueDecimals": metric.precision,
                "valuePrefix": reference_prefix(metric, reference),
                "valueSuffix": reference_suffix(metric, reference),
            },
        }

    def _render_annotation(self, annotation_df, x_axis):
        """
        Group data in the annotation data frame and calculate their positions on the x-axis of the main chart.

        :param annotation_df:
            A data frame containing annotation data.
        :param x_axis:
            The x-axis of the chart.
        :return:
            A list of annotations with label based on the annotation data frame.
        """
        if annotation_df.empty:
            return []

        annotation_alias = annotation_df.columns[0]
        annotation_df[annotation_alias] = annotation_df[annotation_alias].astype(str)

        # Group the annotation frame by concatenating the strings in the annotation column for each index value
        grouped_annotation_df = self._group_annotation_df(
            annotation_df, annotation_alias
        ).to_frame()

        # Calculate the annotation label positions for either timeseries or categories
        annotation_label_positions = self._get_annotation_positions(
            grouped_annotation_df, annotation_alias, x_axis
        )

        annotation = {"labels": []}
        for annotation_position in annotation_label_positions:
            label = {
                "point": {"x": annotation_position["position"], "xAxis": 0},
                "text": annotation_position["label"],
            }

            annotation["labels"].append(label)

        return [annotation]

    def _get_annotation_positions(self, df, dimension_alias, axis):
        if axis["type"] == "datetime":
            return self._get_timeseries_positions(df, dimension_alias)

        return self._get_category_positions(df, dimension_alias, axis)

    @staticmethod
    def _group_annotation_df(annotation_df, annotation_field_name):
        def concatenate_row_values(df):
            df.fillna("None", inplace=True)
            return ", ".join(df)

        return annotation_df.groupby(annotation_df.index.names)[
            annotation_field_name
        ].apply(lambda x: concatenate_row_values(x))

    @staticmethod
    def _get_timeseries_positions(df, dimension_alias):
        timeseries_positions = []

        for dimensions, dimension_value in df[dimension_alias].iteritems():
            datetime = utils.wrap_list(dimensions)[0]

            timeseries_positions.append(
                {"position": formats.date_as_millis(datetime), "label": dimension_value}
            )

        return timeseries_positions

    @staticmethod
    def _get_category_positions(df, dimension_alias, axis):
        dimension_positions = []

        category_positions = {
            category: index for index, category in enumerate(axis["categories"])
        }

        for dimensions, dimension_value in df[dimension_alias].iteritems():
            category_label = utils.wrap_list(dimensions)[0]

            dimension_positions.append(
                {
                    "position": category_positions[category_label],
                    "label": dimension_value,
                }
            )

        return dimension_positions

    @staticmethod
    def _remove_date_totals(data_frame):
        """
        This function filters the totals value for the date/time dimension from the result set. There is no way to
        represent this value on a chart so it is just removed.

        :param data_frame:
        :return:
        """
        if isinstance(data_frame.index, pd.MultiIndex):
            first_index = data_frame.index.get_level_values(0)
            if isinstance(first_index, pd.DatetimeIndex):
                index_slice = first_index < TS_UPPER_BOUND
                return data_frame.loc[index_slice, :]

        elif isinstance(data_frame.index, pd.DatetimeIndex):
            return data_frame[data_frame.index < TS_UPPER_BOUND]

        return data_frame

    @staticmethod
    def _group_by_series(data_frame):
        if len(data_frame) == 0 or not isinstance(data_frame.index, pd.MultiIndex):
            return [([], data_frame)]

        levels = data_frame.index.names[1:]
        return data_frame.groupby(level=levels, sort=False)

    @staticmethod
    def _format_dimension_values(dimensions, dimension_values):
        return ", ".join(
            str.strip(formats.display_value(value, dimension) or str(value))
            for value, dimension in zip(dimension_values, dimensions)
        )



from collections import defaultdict, namedtuple
from typing import TYPE_CHECKING, Union

from pypika.terms import Field as PyPikaField
from toposort import CircularDependencyError, toposort_flatten

from fireant.dataset.filters import Filter
from fireant.dataset.intervals import DATETIME_INTERVALS, DatetimeInterval
from fireant.dataset.modifiers import OmitFromRollup, Rollup
from fireant.dataset.operations import Share
from fireant.exceptions import DataSetException
from fireant.utils import groupby, ordered_distinct_list, ordered_distinct_list_by_attr

if TYPE_CHECKING:
    from fireant import Field

class MissingTableJoinException(DataSetException):
    pass


class CircularJoinsException(DataSetException):
    pass


ReferenceGroup = namedtuple("ReferenceGroup", ("dimension", "time_unit", "intervals"))


def _get_field_definition(field: Union['Field', 'Filter']) -> PyPikaField:
    if field.is_wrapped:
        return getattr(field.definition, 'definition', None)

    return getattr(field, 'definition', None)


def find_required_tables_to_join(elements, base_table):
    """
    Collect all the tables required for a given list of dataset elements.  This looks through the definition and
    display_definition attributes of all elements and

    This looks through the metrics, dimensions, and filter included in this dataset query. It also checks both the
    definition
    field of each element as well as the display definition for Unique Dimensions.

    :return:
        A collection of tables required to execute a query,
    """
    return ordered_distinct_list(
        [
            table
            for element in elements
            # Need extra for-loop to incl. the `display_definition` from `UniqueDimension`
            for attr in [_get_field_definition(element)]
            # ... but then filter Nones since most elements do not have `display_definition`
            if attr is not None
            for table in attr.tables_
            # Omit the base table from this list
            if base_table != table
        ]
    )


def find_joins_for_tables(joins, base_table, required_tables):
    """
    Given a set of tables required for a dataset query, this function finds the joins required for the query and
    sorts them topologically.

    :return:
        A list of joins in the order that they must be joined to the query.
    :raises:
        MissingTableJoinException - If a table is required but there is no join for that table
        CircularJoinsException - If there is a circular dependency between two or more joins
    """
    dependencies = defaultdict(set)
    slicer_joins = {join.table: join for join in joins}

    while required_tables:
        table = required_tables.pop()

        if table not in slicer_joins:
            raise MissingTableJoinException(
                "Could not find a join for table {}".format(str(table))
            )

        join = slicer_joins[table]
        tables_required_for_join = set(join.criterion.tables_) - {
            base_table,
            join.table,
        }

        dependencies[join] |= {
            slicer_joins[table] for table in tables_required_for_join
        }
        required_tables += tables_required_for_join - {d.table for d in dependencies}

    try:
        return toposort_flatten(dependencies, sort=True)
    except CircularDependencyError as e:
        raise CircularJoinsException(str(e))


def find_metrics_for_widgets(widgets):
    """
    :return:
        an ordered, distinct list of metrics used in all widgets as part of this query.
    """
    return ordered_distinct_list_by_attr(
        [metric for widget in widgets for metric in widget.metrics]
    )


def find_operations_for_widgets(widgets):
    """
    :return:
        an ordered, distinct list of metrics used in all widgets as part of this query.
    """
    return ordered_distinct_list_by_attr(
        [operation for widget in widgets for operation in widget.operations]
    )


def find_dataset_fields(metrics):
    """
    Given a list of metrics used in widgets from a dataset blender query, this function returns a list of metrics that
    are from a dataset. Concretely, this means that if a dataset blender has a metric built on dataset metrics, then
    this will replace that metric with the metrics from the dataset.
    """
    from fireant.dataset.fields import Field

    return [
        field or metric
        for metric in metrics
        for field in metric.definition.find_(Field)
    ]


def find_share_operations(operations):
    return [operation for operation in operations if isinstance(operation, Share)]


def find_share_dimensions(dimensions, operations):
    """
    Returns a subset list of dimensions from the list of dimensions that are used as the over-dimension in share
    operations.

    :param dimensions:
    :param operations:
    :return:
    """
    dimension_map = {dimension.alias: dimension for dimension in dimensions}

    return [
        dimension_map[operation.over.alias]
        for operation in operations
        if isinstance(operation, Share)
        and operation.over is not None
        and operation.over.alias in dimension_map
    ]


def find_totals_dimensions(dimensions, share_dimensions):
    """
    :param dimensions:
    :param share_dimensions:
    :return:
        an list of all dimension field in the list argument `dimensions` which have the `Rollup` modifier applied to
        them or are used as a basis for a share metric.
    """
    share_dimension_aliases = {d.alias for d in share_dimensions}
    return [
        dimension
        for dimension in dimensions
        if isinstance(dimension, Rollup) or dimension.alias in share_dimension_aliases
    ]


def find_filters_for_totals(filters):
    """
    :param filters:
    :return:
        a list of filters that should be applied to totals queries. This removes any filters from the list that have
        the `OmitFromRollup` modifier applied to them.
    """
    return [fltr for fltr in filters if not isinstance(fltr, OmitFromRollup)]


def find_field_in_modified_field(field):
    """
    Returns the field from a modified field argument (or just the field argument if it is not modified). A modified
    field represents either a wrapped dimension (e.g. DatetimeInterval) or a field wrapped in a filter.
    """
    modified_field = field
    while hasattr(modified_field, "dimension"):
        modified_field = modified_field.dimension
    while hasattr(modified_field, "field"):
        modified_field = modified_field.field
    return modified_field


interval_weekdays = {
    "month": ("week", 4),
    "quarter": ("week", 4 * 3),
    "year": ("week", 4 * 13),
}


def find_and_group_references_for_dimensions(dimensions, references):
    """
    Finds all of the references for dimensions and groups them by dimension, interval unit, number of intervals.

    This structure reflects how the references need to be joined to the dataset query. References of the same
    type (WoW, WoW.delta, WoW.delta_percent) can share a join query.

    :param dimensions:
    :param references:

    :return:
        An `OrderedDict` where the keys are 3-item tuples consisting of "Dimension, interval unit, # of intervals.

        .. code-block:: python

            Example
            {
                (Dimension(date_1), 'weeks', 1): [WoW, WoW.delta],
                (Dimension(date_1), 'years', 1): [YoY],
                (Dimension(date_7), 'days', 1): [DoD, DoD.delta_percent],
            }
    """
    align_weekdays = (
        dimensions
        and isinstance(dimensions[0], DatetimeInterval)
        and -1 < DATETIME_INTERVALS.index(dimensions[0].interval_key) < 3
    )

    def get_dimension_time_unit_and_interval(reference):
        defaults = (reference.time_unit, 1)
        time_unit, interval_muliplier = (
            interval_weekdays.get(reference.time_unit, defaults)
            if align_weekdays
            else defaults
        )

        field = find_field_in_modified_field(reference.field)
        return field, time_unit, interval_muliplier * reference.interval

    distinct_references = ordered_distinct_list(references)
    return groupby(distinct_references, get_dimension_time_unit_and_interval)



import itertools

from fireant import utils
from fireant.reference_helpers import (
    reference_alias,
    reference_label,
)
from .base import TransformableWidget
from .chart_base import ChartWidget

MAP_SERIES_TO_PLOT_FUNC = {
    ChartWidget.LineSeries: 'line',
    ChartWidget.AreaSeries: 'area',
    ChartWidget.AreaStackedSeries: 'area',
    ChartWidget.AreaPercentageSeries: 'area',
    ChartWidget.PieSeries: 'pie',
    ChartWidget.BarSeries: 'bar',
    ChartWidget.StackedBarSeries: 'bar',
    ChartWidget.ColumnSeries: 'bar',
    ChartWidget.StackedColumnSeries: 'bar',
}


class Matplotlib(ChartWidget, TransformableWidget):
    def __init__(self, title=None):
        super(Matplotlib, self).__init__()
        self.title = title

    def transform(self, data_frame, dimensions, references, annotation_frame=None):
        import matplotlib.pyplot as plt
        result_df = data_frame.copy()

        hide_dimensions = {
            dimension for dimension in dimensions if dimension.fetch_only
        }
        self.hide_data_frame_indexes(result_df, hide_dimensions)

        n_axes = len(self.items)
        figsize = (14, 5 * n_axes)
        fig, plt_axes = plt.subplots(n_axes,
                                     sharex='row',
                                     figsize=figsize)
        fig.suptitle(self.title)

        if not hasattr(plt_axes, '__iter__'):
            plt_axes = (plt_axes,)

        colors = itertools.cycle('bgrcmyk')
        for axis, plt_axis in zip(self.items, plt_axes):
            for series in axis:
                series_color = next(colors)

                linestyles = itertools.cycle(['-', '--', '-.', ':'])
                for reference in [None] + references:
                    metric = series.metric
                    f_metric_key = utils.alias_selector(reference_alias(metric, reference))
                    f_metric_label = reference_label(metric, reference)

                    plot = self.get_plot_func_for_series_type(result_df[f_metric_key], f_metric_label, series)
                    plot(ax=plt_axis,
                         label=axis.label,
                         color=series_color,
                         stacked=series.stacking is not None,
                         linestyle=next(linestyles)) \
                        .legend(loc='center left',
                                bbox_to_anchor=(1, 0.5))

        return plt_axes

    @staticmethod
    def get_plot_func_for_series_type(pd_series, label, chart_series):
        pd_series.name = label
        plot = pd_series.plot
        plot_func_name = MAP_SERIES_TO_PLOT_FUNC[type(chart_series)]
        plot_func = getattr(plot, plot_func_name)
        return plot_func



from pypika import RedshiftQuery

from .postgresql import PostgreSQLDatabase


class RedshiftDatabase(PostgreSQLDatabase):
    """
    Redshift client that uses the psycopg module.
    """

    # The pypika query class to use for constructing queries
    query_cls = RedshiftQuery

    def __init__(self, host='localhost', port=5439, database=None,
                 user=None, password=None, **kwargs):
        super(RedshiftDatabase, self).__init__(host, port, database, user, password, **kwargs)



from functools import partial
from typing import Iterable

import pandas as pd

from fireant import formats
from fireant.dataset.fields import Field
from fireant.utils import alias_selector, wrap_list
from .base import ReferenceItem, TransformableWidget

HARD_MAX_COLUMNS = 24


class Pandas(TransformableWidget):
    def __init__(
        self,
        metric: Field,
        *metrics: Iterable[Field],
        pivot=(),
        hide=(),
        transpose=False,
        sort=None,
        ascending=None,
        max_columns=None
    ):
        super(Pandas, self).__init__(metric, *metrics)
        self.pivot = pivot
        self.hide = hide
        self.transpose = transpose
        self.sort = sort
        self.ascending = ascending
        self.max_columns = (
            min(max_columns, HARD_MAX_COLUMNS)
            if max_columns is not None
            else HARD_MAX_COLUMNS
        )

    def transform(
        self,
        data_frame,
        dimensions,
        references,
        annotation_frame=None,
        use_raw_values=False,
    ):
        """
        WRITEME

        :param data_frame:
        :param dimensions:
        :param references:
        :param annotation_frame:
            A data frame containing the annotation data.
        :param use_raw_values:
            Don't add prefix or postfix to values.
        """
        result_df = data_frame.copy()

        dimension_aliases = [
            alias_selector(dimension.alias) for dimension in dimensions
        ]

        items = [
            item if reference is None else ReferenceItem(item, reference)
            for item in self.items
            for reference in [None] + references
        ]

        if isinstance(result_df.index, pd.MultiIndex):
            result_df = result_df.reorder_levels(dimension_aliases)

        result_df = result_df[[alias_selector(item.alias) for item in items]]

        hide_dimensions = set(self.hide) | {
            dimension for dimension in dimensions if dimension.fetch_only
        }
        self.hide_data_frame_indexes(result_df, hide_dimensions)

        hide_aliases = {
            dimension.alias for dimension in hide_dimensions
        }

        if dimensions:
            result_df.index.names = [
                dimension.label or dimension.alias
                for dimension in dimensions
                if dimension.alias not in hide_aliases
            ]

        result_df.columns = pd.Index([item.label for item in items], name="Metrics")

        pivot_dimensions = [
            dimension.label or dimension.alias
            for dimension in self.pivot
            if dimension.alias not in hide_aliases
        ]
        result_df, _, _ = self.pivot_data_frame(result_df, pivot_dimensions, self.transpose)
        return self.add_formatting(dimensions, items, result_df, use_raw_values).fillna(
            value=formats.BLANK_VALUE
        )

    @staticmethod
    def _should_data_frame_be_transformed(data_frame, pivot_dimensions, transpose):
        if not pivot_dimensions and not transpose:
            return False

        if transpose and len(pivot_dimensions) == len(data_frame.index.names):
            # Pivot and transpose cancel each other out
            return False

        return True

    def pivot_data_frame(self, data_frame, pivot_dimensions, transpose):
        """
        Pivot and transpose the data frame. Dimensions including in the `pivot` arg will be unshifted to columns. If
        `transpose` is True the data frame will be transposed. If there is only index level in the data frame (ie. one
        dimension), and that dimension is pivoted, then the data frame will just be transposed. If there is a single
        metric in the data frame and at least one dimension pivoted, the metrics column level will be dropped for
        simplicity.

        :param data_frame:
            The result set data frame
        :param pivot_dimensions:
            A list of index aliases for `data_frame` of levels to shift
        :param transpose:
            A boolean true or false whether to transpose the data frame.
        :return:
            Tuple(The shifted/transposed data frame, is_pivoted, is_transposed)
        """
        is_pivoted = False
        is_transposed = False

        if not self._should_data_frame_be_transformed(data_frame, pivot_dimensions, transpose):
            return self.sort_data_frame(data_frame), is_pivoted, is_transposed

        # NOTE: Don't pivot a single dimension data frame. This turns the data frame into a series and pivots the
        # metrics anyway. Instead, transpose the data frame.
        should_transpose_instead_of_pivot = len(pivot_dimensions) == len(data_frame.index.names)

        if pivot_dimensions and not should_transpose_instead_of_pivot:
            data_frame = data_frame.unstack(level=pivot_dimensions)
            is_pivoted = True

        if transpose or should_transpose_instead_of_pivot:
            data_frame = data_frame.transpose()
            is_transposed = True

        # If there are more than one column levels and the last level is a single metric, drop the level
        if isinstance(data_frame.columns, pd.MultiIndex) and 1 == len(
            data_frame.columns.levels[0]
        ):
            data_frame.name = data_frame.columns.levels[0][
                0
            ]  # capture the name of the metrics column
            data_frame.columns = data_frame.columns.droplevel(
                0
            )  # drop the metrics level

        return self.sort_data_frame(data_frame), is_pivoted, is_transposed

    def sort_data_frame(self, data_frame):
        if not self.sort or len(data_frame) == 1:
            # If there are no sort arguments or the data frame is a single row, then no need to sort
            return data_frame

        # reset the index so all columns can be sorted together
        index_names = data_frame.index.names
        unsorted = data_frame.reset_index()
        column_names = list(unsorted.columns)

        ascending = self.ascending if self.ascending is not None else True

        sort = wrap_list(self.sort)
        sort_columns = [
            column_names[column] for column in sort if column < len(column_names)
        ]

        if not sort_columns:
            return data_frame

        # ignore additional values for ascending if they do not align lengthwise with sort_columns
        # Default to the first value or None
        if isinstance(ascending, list) and len(ascending) != len(sort_columns):
            ascending = ascending[0] if len(ascending) > 0 else None

        sorted = unsorted.sort_values(sort_columns, ascending=ascending).set_index(
            index_names
        )

        # Maintain the single metric name
        if hasattr(data_frame, "name"):
            sorted.name = data_frame.name

        return sorted

    def add_formatting(self, dimensions, items, pivot_df, use_raw_values):
        format_df = pivot_df.copy()

        def _get_field_display(item):
            return partial(
                formats.display_value,
                field=item,
                nan_value="",
                null_value="",
                use_raw_value=use_raw_values,
            )

        if (
            self.transpose
            or not self.transpose
            and len(dimensions) == len(self.pivot) > 0
        ):
            for item in items:
                field_display = _get_field_display(item)
                format_df.loc[items[0].label] = format_df.loc[items[0].label].apply(
                    field_display
                )

            return format_df

        if self.pivot and len(items) == 1:
            field_display = _get_field_display(items[0])
            format_df = format_df.applymap(field_display)
            return format_df

        for item in items:
            key = item.label
            field_display = _get_field_display(item)
            format_df[key] = (
                format_df[key].apply(field_display)
                if isinstance(format_df[key], pd.Series)
                else format_df[key].applymap(field_display)
            )

        return format_df
